"""MQTT Client interface for pypglab."""

from __future__ import annotations

from collections.abc import Callable, Coroutine
from typing import Any

# define a mqtt subscribe callback ... the input argument are payload and the topic string
Subcribe_CallBack = Callable[[str, str], None]

Sub_State = dict[str, Any]
Sub_States = dict[int, Sub_State]


class Client:
    """The client interface to mqtt."""

    def __init__(
        self,
        publish: Callable[
            [str, str, int | None, bool | None], Coroutine[Any, Any, dict]
        ],
        subscribe: Callable[
            [Sub_State, str, Subcribe_CallBack], Coroutine[Any, Any, dict]
        ],
        unsubscribe: Callable[[Sub_State], Coroutine[Any, Any, dict]],
    ) -> None:
        """Initiliaze."""

        self._publish = publish
        self._subscribe = subscribe
        self._unsubscribe = unsubscribe
        self._substates: Sub_States = {}

    async def publish(
        self, topic: str, payload: str, qos: int | None = 0, retain: bool | None = False
    ) -> None:
        """Publis a MQTT message."""
        if self._publish:
            await self._publish(topic, payload, qos, retain)

    async def subscribe(
        self, unique_id: int, topic: str, callback: Subcribe_CallBack
    ) -> None:
        """Subscriobe to a MQTT topic."""

        if self._subscribe:
            # get the hash value from the topic
            hash_value = hash((unique_id, topic))

            # from all subribed topic states create/get one that match the hash
            substate = self._substates.get(hash_value)

            # call the mqtt client and get an substate update
            substate = await self._subscribe(substate, topic, callback)

            # save the sub state
            self._substates[hash_value] = substate

    async def unsubscribe(self, unique_id: int, topic: str) -> None:
        """Unsubscribe from a MQTT topics."""

        if self._unsubscribe:
            # get the substate from the topic
            hash_value = hash((unique_id, topic))
            substate = self._substates.get(hash_value)

            # call the mqtt client to unsubscribe
            substate = await self._unsubscribe(substate)

            # update the substate
            self._substates[hash_value] = substate



"""Sensor for pypglab."""

import json
from typing import Any, cast

from .const import ENTITY_SENSOR, SENSOR_REBOOT_TIME, SENSOR_TEMPERATURE, SENSOR_VOLTAGE
from .entity import Entity
from .mqtt import Client


def SensorDefaultValue(sensor_type: str) -> Any:
    """Return a default value for a sensor type."""
    if sensor_type == SENSOR_TEMPERATURE:
        return 0
    if sensor_type == SENSOR_VOLTAGE:
        return 0
    if sensor_type == SENSOR_REBOOT_TIME:
        return 0
    return None


def SensorValueCast(sensor_type: str, value: Any) -> Any:
    """Cast a value to the specific type of the sensor."""
    if sensor_type == SENSOR_TEMPERATURE:
        return cast(int, value)
    if sensor_type == SENSOR_VOLTAGE:
        return cast(int, value)    
    if sensor_type == SENSOR_REBOOT_TIME:
        return cast(int, value)
    return None


class Sensor(Entity):
    """It's a PG LAB Electronics sensor."""

    def __init__(
        self,
        device_id: str,
        device_name: str,
        config: [str],
        mqtt: Client,
    ) -> None:
        """Initialize."""
        super().__init__(device_id, device_name, 0, ENTITY_SENSOR, mqtt)

        # initialize all sensor value
        self._state: dict = {}
        for sensor_type in config:
            self._state[sensor_type] = SensorDefaultValue(sensor_type)

    def _getSensorValue(self, sensor_type: str, values: dict) -> Any:
        if sensor_type in values:
            return SensorValueCast(sensor_type, values[sensor_type])

        return None

    def status_change_received(self, payload: str) -> None:
        """Call to notify a new status change."""
        values = json.loads(payload)

        for s in self._state:
            newValue = self._getSensorValue(s, values)
            if newValue:
                self._state[s] = newValue

    @property
    def state(self) -> dict:
        """Get sensor status."""
        return self._state

    @property
    def size(self) -> int:
        """Return the number of stored sensor value."""
        return len(self._state)


async def CreateSensor(
    device_id: str, device_name: str, config: [str], mqtt: Client
) -> Sensor:
    """Create and initialize a PG LAB relay instance."""

    sensor = Sensor(device_id, device_name, config, mqtt)
    await sensor.subscribe_topics()
    return sensor



"""Shutter for pypglab."""

from .const import (
    ENTITY_SHUTTER,
    SHUTTER_CMD_CLOSE,
    SHUTTER_CMD_OPEN,
    SHUTTER_CMD_STOP,
    SHUTTER_STATE_CLOSED,
    SHUTTER_STATE_CLOSING,
    SHUTTER_STATE_OPEN,
    SHUTTER_STATE_OPENING,
)
from .entity import Entity
from .mqtt import Client


class Shutter(Entity):
    """It's a PG LAB Electronics shutter."""

    STATE_UNKNOWN = 0
    STATE_OPENING = 1
    STATE_OPEN = 2
    STATE_CLOSING = 3
    STATE_CLOSED = 4

    def __init__(
        self,
        device_id: str,
        device_name: str,
        index: int,
        mqtt: Client,
    ) -> None:
        """Initialize."""
        super().__init__(device_id, device_name, index, ENTITY_SHUTTER, mqtt)

        self._state = None

    def status_change_received(self, payload: str) -> None:
        """Call o notify a new status change."""
        if payload == SHUTTER_STATE_OPENING:
            self._state = Shutter.STATE_OPENING
        elif payload == SHUTTER_STATE_OPEN:
            self._state = Shutter.STATE_OPEN
        elif payload == SHUTTER_STATE_CLOSING:
            self._state = Shutter.STATE_CLOSING
        elif payload == SHUTTER_STATE_CLOSED:
            self._state = Shutter.STATE_CLOSED

    async def open(self) -> None:
        """Open the shutter."""
        await self.set_state(SHUTTER_CMD_OPEN)

    async def close(self) -> None:
        """Close the sutter."""
        await self.set_state(SHUTTER_CMD_CLOSE)

    async def stop(self) -> None:
        """Stop the shutter."""
        await self.set_state(SHUTTER_CMD_STOP)

    @property
    def state(self) -> int:
        """Get shutter status."""
        return self._state


async def CreateShutter(
    device_id: str, device_name: str, index: int, mqtt: Client
) -> Shutter:
    """Create and initialize a PG LAB shutter instance."""

    shutter = Shutter(device_id, device_name, index, mqtt)
    await shutter.subscribe_topics()
    return shutter



import paho.mqtt.client as mqtt

from pypglab.mqtt import Client, Sub_State, Subcribe_CallBack

MQTT_SERVER = "localhost"
MQTT_PORT = 1883
MQTT_USERNAME = "pierluigi"
MQTT_PASSWORD = "password"

PGLAB_DISCOVERY_TOPIC = "pglab/discovery"


class MQTT:
    """MQTT Client class to comunicate with PG LAB device using paho MQTT library."""
    
    def __init__(self) -> None:
        """Initiliaze."""

        def on_connect(client, userdata, flags, rc):
            """The callback for when the client receives a CONNACK response from the server."""
            self._connected = True


        def on_disconnect(client, userdata, rc):
            self._connected = False
        
        def on_message(client, userdata, msg):
            """Callback when a message has been received on a topic that the client subscribes."""

            if msg.topic.startswith(PGLAB_DISCOVERY_TOPIC):
                self._discovery.append( (msg.topic, msg.payload) )
                return 
            
            if userdata is None:
                # this should never happen
                return

            # do the callback
            callback = userdata.get(msg.topic)
            if callback:
                payload = msg.payload.decode("utf-8")
                callback(msg.topic, payload)


        self._connected = False
        self._discovery = []
        self._subscribe_callback:dict = {}
        self._mqtt_client = mqtt.Client()
        self._mqtt_client.on_connect = on_connect
        self._mqtt_client.on_disconnect = on_disconnect
        self._mqtt_client.on_message = on_message
        self._mqtt_client.user_data_set(self._subscribe_callback)
        self._mqtt_client.username_pw_set(MQTT_USERNAME, MQTT_PASSWORD)
        # create a mqtt client for pglab used for pglab python module
        self._pglab_mqtt_client : Client = None

    def setup(self)->None:

        async def mqtt_publish(
            topic: str, payload: str, qos: int | None = 0, retain: bool | None = False
        ) -> None:
            self._mqtt_client.publish(topic, payload, qos, retain)

        async def mqtt_subscribe(
            sub_state: Sub_State, topic: str, callback_func: Subcribe_CallBack
        ) -> Sub_State:
            
            self._subscribe_callback[topic] = callback_func
            self._mqtt_client.subscribe(topic, qos=0)

            new_substate = {}
            new_substate[topic] = callback_func

            self._subscribe_callback[topic] = callback_func

            return new_substate

        async def mqtt_unsubscribe(sub_state: Sub_State) -> None:
            for topic in sub_state:
                self._mqtt_client.unsubscribe(topic)

        # connect with MQTT broker
        self._mqtt_client.connect(MQTT_SERVER, MQTT_PORT, 60)

        # setup the MQTT PGLAB client interface
        self._pglab_mqtt_client = Client(mqtt_publish, mqtt_subscribe, mqtt_unsubscribe)

        # setup to PGLAB discovery message
        self._mqtt_client.subscribe(PGLAB_DISCOVERY_TOPIC + "/#")

    def start(self) -> None:
        """Start the MQTT Loop."""
        self._mqtt_client.loop_start()

    def stop(self) -> None:
        """Stop the MQTT Loop."""
        self._mqtt_client.loop_stop()

    @property
    def discovery(self) :
        return self._discovery
    
    @property
    def pglab_mqtt_client(self):
        return self._pglab_mqtt_client
    
    @property
    def connected(self):
        return self._connected





"""Relay for pypglab."""

from .const import ENTITY_RELAY, RELAY_STATE_OFF, RELAY_STATE_ON
from .entity import Entity
from .mqtt import Client


class Relay(Entity):
    """It's a PG LAB Electronics relay."""

    def __init__(
        self,
        device_id: str,
        device_name: str,
        index: int,
        mqtt: Client,
    ) -> None:
        """Initialize."""
        super().__init__(device_id, device_name, index, ENTITY_RELAY, mqtt)

        self._state: bool = None

    def status_change_received(self, payload: str) -> None:
        """Call o notify a new status change."""
        self._state = True if payload == RELAY_STATE_ON else False

    async def __set_state(self, state: bool) -> None:
        """Turn the relay on or off."""
        payload = RELAY_STATE_ON if state else RELAY_STATE_OFF
        await super().set_state(payload)

    async def turn_on(self) -> None:
        """Turn the relay on."""
        await self.__set_state(True)

    async def turn_off(self) -> None:
        """Turn the relay on."""
        await self.__set_state(False)

    @property
    def state(self) -> bool:
        """Get relay status."""
        return self._state


async def CreateRelay(
    device_id: str, device_name: str, index: int, mqtt: Client
) -> Relay:
    """Create and initialize a PG LAB relay instance."""

    relay = Relay(device_id, device_name, index, mqtt)
    await relay.subscribe_topics()
    return relay



""" A device for pypglab."""

from __future__ import annotations

from typing import cast
from voluptuous import Schema, ALLOW_EXTRA, REMOVE_EXTRA, MultipleInvalid

from . import validation as cv

from .const import (
    E_BOARD, 
    E_SWITCH, 
    LOGGER, 
    PGLAB_DEVICE_TYPES, 
    SENSOR_CONFIG, 
    CONFIG_MAC, 
    CONFIG_IP, 
    CONFIG_ID, 
    CONFIG_NAME, 
    CONFIG_PARAMETERS,
    CONFIG_TYPE, 
    CONFIG_MANUFACTOR, 
    CONFIG_HARDWARE_VERSION, 
    CONFIG_FIRMWARE_VERSION,
    CONFIG_EBOARD_SHUTTERS,
    CONFIG_EBOARD_BOARDS,
)

from .mqtt import Client
from .relay import CreateRelay
from .sensor import CreateSensor
from .shutter import CreateShutter


PGLAB_DISCOVERY_SCHEMA = Schema(
    {
        CONFIG_MAC: cv.macaddrs,
        CONFIG_IP: cv.ipaddrs,
        CONFIG_ID: cv.string,
        CONFIG_NAME: cv.string,
        CONFIG_TYPE: cv.devicetype,
        CONFIG_MANUFACTOR: cv.manufacturer,
        CONFIG_HARDWARE_VERSION: cv.version,
        CONFIG_FIRMWARE_VERSION: cv.version,
    },
    required=True,
    extra=ALLOW_EXTRA,
)

PGLAB_EBOARD_PARAMETERS = Schema (
    { 
        CONFIG_PARAMETERS: 
        {
            CONFIG_EBOARD_SHUTTERS: cv.shutter,
            CONFIG_EBOARD_BOARDS: cv.boards,
        }
    },
    required=True,
    extra=REMOVE_EXTRA,
)

class Device:
    """The class represent a generic PG LAB device."""

    def __init__(self) -> None:
        """Initiliaze."""

        # device ip address
        self._ip = None

        # the device mac address
        self._mac = None

        # device unique id along all PG Lab devices
        self._id = None

        # device friendly name for mqtt connection
        self._name = None

        # device name type
        self._type = None

        # the manufactor of the device
        self._manufacturer = None

        # the hardware version
        self._hardware_version = None

        # the firmware version
        self._firmware_version = None

        # specific parameters of the device
        self._parameters = None

        # prepare an array of available relays
        self._relays = []

        # prepare an array of available shutters
        self._shutters = []

        # prepare the sensors of the device
        self._sensors = None

        # internal configuration hash
        self._hash = hash(
            (
                self._id,
                self._name,
            )
        )

    async def config(self, mqtt: Client, config: dict) -> bool:
        """Perform internal configuration."""

        # validate config message
        try:
            config = PGLAB_DISCOVERY_SCHEMA(config)
        except MultipleInvalid as e:
            LOGGER.warning("Invalid discovey message (%s)", e)
            return False

        self._mac = config[CONFIG_MAC]
        self._ip = config[CONFIG_IP]
        self._id = config[CONFIG_ID]
        self._name = config[CONFIG_NAME]
        self._type = config[CONFIG_TYPE]
        self._manufacturer = config[CONFIG_MANUFACTOR]
        self._hardware_version = config[CONFIG_HARDWARE_VERSION]
        self._firmware_version = config[CONFIG_FIRMWARE_VERSION]

        # update the device hash
        self._hash = hash((self._id, self._name, self._mac))

        # initialize the specific type of device
        if self.is_eboard:

            # validate e-board parameters
            try:
                eboard_config = PGLAB_EBOARD_PARAMETERS(config)                
            except MultipleInvalid as e:
                LOGGER.warning("Invalid E-Board parameters message (%s)", e)
                return False

            # save the e-board parameters
            self._parameters = eboard_config[CONFIG_PARAMETERS]

            # update the device hash with specific e-board configuration
            self._hash = hash(
                (
                    self._hash,
                    self._parameters.get(CONFIG_EBOARD_SHUTTERS),
                    self._parameters.get(CONFIG_EBOARD_BOARDS),
                )
            )

            shutters = self._parameters.get(CONFIG_EBOARD_SHUTTERS)

            # prepare all shutters
            for index in range(0, shutters):
                if await self.is_relay_connected(index * 2):
                    shutter = await CreateShutter(self._id, self._name, index, mqtt)
                    self._shutters.append(shutter)

            # prepare all relays
            for index in range(2 * shutters, 64):
                if await self.is_relay_connected(index):
                    relay = await CreateRelay(self._id, self._name, index, mqtt)
                    self._relays.append(relay)

            # prepare the sensor
            self._sensors = await CreateSensor(
                self._id, self._name, SENSOR_CONFIG[self._type], mqtt
            )
        
        return True

    async def is_relay_connected(self, index):
        """Return if a specific relay is connected to e-board and available to use."""
        if self.is_eboard:
            if index < 0 or index > 63:
                return False

            boards = self._parameters[CONFIG_EBOARD_BOARDS]
            bi = index // 8

            return boards[bi] == "1"
        else:
            return False

    @property
    def is_eboard(self):
        """Return true if it is a E-Board device."""
        return self.type == PGLAB_DEVICE_TYPES[E_BOARD]

    @property
    def is_eswitch(self):
        """Return true if it is a E-Switch device."""
        return self.type == PGLAB_DEVICE_TYPES[E_SWITCH]

    @property
    def relays(self):
        """Get the relay array."""
        return self._relays

    @property
    def shutters(self):
        """Get the shutter array."""
        return self._shutters

    @property
    def sensors(self):
        """Get the device sensors ."""
        return self._sensors

    @property
    def ip(self) -> str:
        """Device ip."""
        return cast(str, self._ip)

    @property
    def mac(self) -> str:
        """Device mac."""
        return cast(str, self._mac)

    @property
    def id(self) -> str:
        """Device unique id."""
        return cast(str, self._id)

    @property
    def name(self) -> str:
        """Device friendly name."""
        return cast(str, self._name)

    @property
    def type(self) -> str:
        """Device type name (E-Board, E-Switch etc...)."""
        return cast(str, self._type)

    @property
    def manufactor(self) -> str:
        """Device manufacturer."""
        return cast(str, self._manufacturer)

    @property
    def hardware_version(self) -> str:
        """Device hardware version."""
        return cast(str, self._hardware_version)

    @property
    def firmware_version(self) -> str:
        """Device firmware version."""
        return cast(str, self._firmware_version)

    @property
    def hash(self) -> int:
        """Return device HASH."""
        return cast(int, self._hash)




import unittest
import time
import json

from .mqtt import MQTT
from pypglab.device import Device
from pypglab.relay import Relay
from pypglab.shutter import Shutter
from pypglab.sensor import Sensor
from pypglab.const import SENSOR_TEMPERATURE

UNIT_TEST_CONNECTION_TIMEOUT = 10

# be sure that the PG LAB Electronics device shutter opening/closing time is less than the following time
UNIT_TEST_SHUTTER_OPEN_TIMEOUT  = 25
UNIT_TEST_SHUTTER_CLOSE_TIMEOUT  = 25

UNIT_TEST_SENSOR_TIMEOUT = 65

class TestPgLab(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self._mqtt = MQTT()
        self._mqtt.setup()
        self._mqtt.start()

    async def asyncTearDown(self):
        self._mqtt.stop()

    async def test_broker_connection(self):
        """Test MQTT Broker connection."""

        end_time = time.time() + UNIT_TEST_CONNECTION_TIMEOUT
        while time.time() < end_time:
            if self._mqtt.connected:
                break

        self.assertTrue(self._mqtt.connected, "No connection with MQTT broker.")

    async def _get_discovery(self):
        """Get the PG LAB MQTT discovery message."""
        discovery = None
        end_time = time.time() + UNIT_TEST_CONNECTION_TIMEOUT
        while time.time() < end_time:
            if self._mqtt.discovery:
                discovery = self._mqtt.discovery
                break

        self.assertTrue(len(discovery) > 0, "No PG LAB Electronics Devices are connected to the MQTT broker.")        
        return discovery
    
    async def _relay_togle(self, r: Relay):
        state_update_changed : bool = False

        def state_changed(payload:str)->None:
            nonlocal state_update_changed
            state_update_changed = True
        
        r.add_state_update(state_changed)

        # be sure that the relay is off
        await r.turn_off()
        time.sleep(1)

        await r.turn_on()
        time.sleep(1)
        self.assertTrue(r.state, "Unexpeted relay state.")

        await r.turn_off()
        time.sleep(1)
        self.assertTrue( not r.state, "Unexpeted relay state.")

        self.assertTrue(state_update_changed, "Relay Status update don't received.")


    async def _shutter_togle(self, s: Shutter):
        state_update_changed : bool = False

        def state_changed(payload:str)->None:
            nonlocal state_update_changed
            state_update_changed = True
        
        s.add_state_update(state_changed)

        # the test suppose that the shutter opening/closing time is 
        # bigger than 10 second

        # at this moment the state of the shutter is unknow
        # try to open the shutter for 2 seconds .. so we are sure that the shutter is not close
        await s.open()
        time.sleep(2)

        self.assertTrue( s.state is not Shutter.STATE_CLOSED, "Unexpeted shuttur state")

        # close the shutter and wait enough time to be sure the shutter is fully close
        await s.close()
        time.sleep(UNIT_TEST_SHUTTER_CLOSE_TIMEOUT)

        self.assertTrue( s.state is Shutter.STATE_CLOSED, "Unexpeted shuttur state")

        # we are sure that the shutte is fully close
        # open the shutter and check the state after 1 second
        await s.open()
        time.sleep(1)

        self.assertTrue( s.state is Shutter.STATE_OPENING, "Unexpeted shuttur state")

        # the shutter is opening ... stop and check the state
        await s.stop()
        time.sleep(1)
        self.assertTrue(s.state is Shutter.STATE_OPEN, "Unexpeted shuttur state")
        self.assertTrue(state_update_changed, "Shutter Status update don't received.")

    async def _sensor_state(self, sensor: Sensor):
            
        state_update_changed : bool = False

        def state_changed(payload:str)->None:
            nonlocal state_update_changed
            state_update_changed = True
        
        sensor.add_state_update(state_changed)

        # wait for sensor update
        timeout = time.time() + UNIT_TEST_SENSOR_TIMEOUT
        while (not state_update_changed) and (time.time() < timeout):
            time.sleep(1)

        self.assertTrue(state_update_changed, "Sensor state update don't received.")
        self.assertTrue(sensor.state and len(sensor.state) > 0, "Sensor state unexpected state")

        # check if the cpu temperature value is been received... it should be available in every device
        self.assertTrue(sensor.state.get(SENSOR_TEMPERATURE), "Sensor Temperature not available")
        self.assertTrue(sensor.state.get(SENSOR_TEMPERATURE)>0, "Sensor Temperature unexpected value")

    async def _create_device(self, discovery) -> Device:
        pglab_device = Device()

        # get the discovery message
        discovery_topic, discovery_payload = discovery

        # decode the paylod from json
        discovery_payload = json.loads(discovery_payload)

        # try to configure the pglab_device
        bres = await pglab_device.config(self._mqtt.pglab_mqtt_client, discovery_payload)

        self.assertTrue( bres, "Error during PG LAB Electronics device configuration.")

        return pglab_device
    

    async def test_discovery(self):
        """Test if a PG LAB Electronics device is connected to the broker and publish the discovery message."""
        await self._get_discovery()

    async def test_device(self):
        """Tesf if it's possible to creare a PG LAB Electronics device with the discovery message."""
        pglab_discovery = await self._get_discovery()

        for discovery in pglab_discovery:        
            pglab_device = await self._create_device(discovery)

            if pglab_device.is_eboard:
                self.assertTrue( len(pglab_device.relays) > 0 or len(pglab_device.shutters) > 0, "No shutters or relays available.")

    async def test_relay(self):
        """Tesf if change status on relay status change."""
        pglab_discovery = await self._get_discovery()

        for discovery in pglab_discovery:        
            pglab_device = await self._create_device(discovery)

            if len(pglab_device.relays) == 0:
                return
            
            for r in pglab_device.relays:
                await self._relay_togle(r)


    async def test_shutter(self):
        """Tesf if change status on relay status change."""
        pglab_discovery = await self._get_discovery()

        for discovery in pglab_discovery:        
            pglab_device = await self._create_device(discovery)

            if len(pglab_device.shutters) == 0:
                return
            
            # test only one shutter because is going to take time
            shutter = pglab_device.shutters[0]
            await self._shutter_togle(shutter)


    async def test_sensors(self):
        """Test device sensor value."""
        pglab_discovery = await self._get_discovery()

        for discovery in pglab_discovery:        
            pglab_device = await self._create_device(discovery)

            await self._sensor_state(pglab_device.sensors)
                

def suite():

    suite = unittest.TestSuite()
    suite.addTest(TestPgLab('test_broker_connection'))
    suite.addTest(TestPgLab('test_discovery'))
    suite.addTest(TestPgLab('test_device'))
    suite.addTest(TestPgLab('test_relay'))
    suite.addTest(TestPgLab('test_shutter'))
    suite.addTest(TestPgLab('test_sensors'))

    return suite




"""Constants used by pypglab."""

import logging
from typing import Final

LOGGER = logging.getLogger(__name__)

PGLAB_VERSION = "0.0.1"

# All devices managed by this python library
E_BOARD: Final = "E-BOARD"
E_RELAY: Final = "E-RELAY"
E_SWITCH: Final = "E-SWITCH"

MANUFACTURER: Final = "PG LAB Electronics"

# The following define the device type name
PGLAB_DEVICE_TYPES = {
    E_BOARD: "E-Board",
    E_RELAY: "E-Relay",
    E_SWITCH: "E-Switch",
}

TOPIC_PGLAB: Final = "pglab"

# Discovery message fields
CONFIG_MAC: Final = "mac"
CONFIG_IP: Final = "ip"
CONFIG_ID: Final = "id"
CONFIG_NAME: Final = "name"
CONFIG_TYPE: Final = "type"
CONFIG_MANUFACTOR: Final = "manufacturer"
CONFIG_HARDWARE_VERSION: Final = "hw"
CONFIG_FIRMWARE_VERSION: Final = "fw"
CONFIG_PARAMETERS: Final = "params"
CONFIG_EBOARD_SHUTTERS: Final = "shutters"
CONFIG_EBOARD_BOARDS: Final = "boards"

# Available entities
ENTITY_RELAY: Final = "relay"
ENTITY_SHUTTER: Final = "shutter"
ENTITY_SWITCH: Final = "switch"
ENTITY_SENSOR: Final = "sensor"

# Relay
RELAY_STATE_ON: Final = "ON"
RELAY_STATE_OFF: Final = "OFF"

# Shutter MQTT state
SHUTTER_STATE_OPENING: Final = "OPENING"
SHUTTER_STATE_OPEN: Final = "OPEN"
SHUTTER_STATE_CLOSING: Final = "CLOSING"
SHUTTER_STATE_CLOSED: Final = "CLOSED"

# Shutter MQTT command
SHUTTER_CMD_OPEN: Final = "OPEN"
SHUTTER_CMD_CLOSE: Final = "CLOSE"
SHUTTER_CMD_STOP: Final = "STOP"

# MQTT topic for set and get state
ENTITY_TOPIC = {
    ENTITY_RELAY: ("set", "state"),
    ENTITY_SHUTTER: ("set", "state"),
    ENTITY_SENSOR: (None, "value"),
}

# Sensors value
SENSOR_TEMPERATURE: Final = "temp"
SENSOR_VOLTAGE: Final = "volt"
SENSOR_REBOOT_TIME: Final = "rtime"

# Sensors configuration for a specific PG LAB device
SENSOR_CONFIG = {
    "E-Board": [SENSOR_TEMPERATURE, SENSOR_VOLTAGE, SENSOR_REBOOT_TIME],
    "E-Relay": [],
    "E-Switch": [],
}



import unittest
from tests.test import suite

# IMPORTANT !!!!! 
# Before to run the UnitTest:
#
# 1) Be sure that a MQTT broker is running and reachable from the unittest machine
#    Connfigure the USERNAME/PASSWORD in ./tests/mqtt.py
# 2) At least one PG LAB Electronics device must be connected to the MQTT broker
# 3) If the device is an E-Board be sure to have it configured to manage shutter and relay.
#    Be sure that the shutter opeing/closing time is less thant the value 
#    UNIT_TEST_SHUTTER_OPEN_TIMEOUT  and  UNIT_TEST_SHUTTER_CLOSE_TIMEOUT  in ./tests/test.py

if __name__ == '__main__':

    runner = unittest.TextTestRunner(failfast=True)
    runner.run(suite())


"""PG LAB Electronics config validation."""
from __future__ import annotations
from typing import Any

import ipaddress
import re
import voluptuous as vol

from . const import PGLAB_DEVICE_TYPES, E_BOARD, E_RELAY, E_SWITCH, MANUFACTURER

def ipaddrs(value: Any)-> str:
    
    if value is None:
        raise vol.Invalid("IP address value is None.")        
    try:
        ipaddress.ip_address(value)
        return str(value)
    except ValueError:
        raise vol.Invalid("IP address is not valid.")
    

def macaddrs(value: Any)-> str:

    if value is None:
        raise vol.Invalid("MAC address value is None.")        

    result = re.match(r"([0-9a-fA-F]{2}[-:]){5}[0-9a-fA-F]{2}$",value)

    if result is None:
        raise vol.Invalid("MAC address is not valid.")        
    
    return str(value)

def devicetype(value: Any)-> str:

    if value is None:
        raise vol.Invalid("Device type value is None.")        

    if value == PGLAB_DEVICE_TYPES[E_BOARD]:
        return str(value)
    elif value == PGLAB_DEVICE_TYPES[E_RELAY]:
        return str(value)
    elif value == PGLAB_DEVICE_TYPES[E_SWITCH]:
        return str(value)

    raise vol.Invalid("Device type is not valid.")        
    
    return None

def version(value: Any)-> str:
    if value is None:
        raise vol.Invalid("Version Number is None.")
    
    result = re.match(r"^([0-9]+)\.([0-9]+)\.([0-9]+)?$",value)
    if result is None:
        raise vol.Invalid("Version Number is not valid.") 
    
    return str(value)

def manufacturer(value: Any)-> str:
    if value is None:
        raise vol.Invalid("Manufacturer string is None.")
    
    if value != MANUFACTURER:
        raise vol.Invalid("Unexpected Manufacturer string is not valid.") 
    return str(value)

def string(value: Any)-> str:
    if value is None:
        raise vol.Invalid("String value is None.")
    
    return str(value)


def shutter(value: Any)-> int:
    if value is None:
        raise vol.Invalid("Shutter number value is None.")

    if value < 0 or value > 32:
        raise vol.Invalid("Unexpected shutter numbers.")
    
    return int(value)

def boards(value:Any)-> str:
    if value is None:
        raise vol.Invalid("Boards connection string is None.")
    
    result = re.match(r"^[0-1]{8}$",value)
    if result is None:
        raise vol.Invalid("Invalid Boards connection string.") 

    return str(value)


import asyncio
import time
import paho.mqtt.client as mqtt

from pypglab.device import Device
from pypglab.mqtt import Client

MQTT_SERVER = "localhost"
MQTT_PORT = 1883
MQTT_USERNAME = "pierluigi"
MQTT_PASSWORD = "password"


async def mqtt_publish( topic: str, payload: str, qos: int | None = 0, retain: bool | None = False) -> None:
    mqtt_client.publish(topic, payload, qos, retain)


mqtt_client = mqtt.Client()
mqtt_client.loop_start()
mqtt_client.username_pw_set(MQTT_USERNAME, MQTT_PASSWORD)
mqtt_client.connect(MQTT_SERVER, MQTT_PORT, 60)

# wait the connection with the broker
time.sleep(1)

pglab_mqtt_client = Client(mqtt_publish, None, None)

config = {
    "ip":"192.168.1.16", 
    "mac":"80:34:28:1B:18:5A", 
    "name":"e-board-office", 
    "hw":"255.255.255", 
    "fw":"255.255.255", 
    "type":"E-Board", 
    "id":"E-Board-DD53AC85", 
    "manufacturer":"PG LAB Electronics", 
    "params":{
        "shutters":3, 
        "boards":"10000000" 
    }
}

pglab_device = Device()
asyncio.run( pglab_device.config(pglab_mqtt_client, config) )

for relay in pglab_device.relays:
    asyncio.run( relay.turn_on() )

mqtt_client.loop_stop()



"""A base class entity for pypglab."""

from collections.abc import Callable

from .const import ENTITY_SENSOR, ENTITY_TOPIC, TOPIC_PGLAB
from .mqtt import Client

State_Update = Callable[[str], None]


class Entity:
    """Base class for PG LAB entities."""

    # define a global id incremented for every instance
    entity_id: int = 0

    def __init__(
        self,
        device_id: str,
        device_name: str,
        entity_id: int,
        entity_type: str,
        mqtt: Client,
    ) -> None:
        """Initialize."""

        # device id connected to this entity
        self._device_id = device_id

        # device name connected to this entity
        self._device_name = device_name

        # entity id, usually a internal index
        self._id = entity_id

        # entity type
        self._type = entity_type

        # mqtt client interface
        self._mqtt = mqtt

        # external status update callback
        self._state_update: [State_Update] = []

        # topic used for change entity status
        if entity_type is ENTITY_SENSOR:
            self._topic = f"{TOPIC_PGLAB}/{self._device_name}/{entity_type}/"
        else:
            self._topic = f"{TOPIC_PGLAB}/{self._device_name}/{entity_type}/{self._id}/"

        # entity hash, it's a uniquie entity identifier
        self._hash: int = hash(
            (Entity.entity_id, self._device_id, self._type, self._id)
        )

        # increment the global entity id
        Entity.entity_id = Entity.entity_id + 1

    def status_change_received(self, payload: str) -> None:
        """Call to notify a new status change.

        To be overwritten by child class.
        """

    def add_state_update(self, state_update: State_Update) -> None:
        """Set a callback to inform about new state."""
        self._state_update.append(state_update)

    async def subscribe_topics(self) -> None:
        """PG LAB Entity subscribe to mqtt relay status changing."""

        def on_message(topic, payload) -> None:
            self.status_change_received(payload)

            for state_update in self._state_update:
                state_update(payload)

        set_cmd, state_cmd = ENTITY_TOPIC[self._type]
        status_update_topic = self._topic + state_cmd
        await self._mqtt.subscribe(self._hash, status_update_topic, on_message)

    async def unsubscribe_topics(self) -> None:
        """Unsubscribe from all MQTT topics."""

        set_cmd, state_cmd = ENTITY_TOPIC[self._type]
        status_update_topic = self._topic + state_cmd
        await self._mqtt.unsubscribe(self._hash, status_update_topic)

    async def set_state(self, payload: str) -> None:
        """Change the entity state."""
        set_cmd, state_cmd = ENTITY_TOPIC[self._type]
        # check if the entity allows to change status
        if set_cmd:
            topic = self._topic + set_cmd
            await self._mqtt.publish(topic, payload)

    @property
    def hash(self) -> int:
        """Return the entity uniquie id hash value."""
        return self._hash

    @property
    def id(self) -> int:
        """Get entity index id."""
        return self._id
